(* Time-stamp: <modified the 29/03/2019 (at 13:33) by Erwan Jahier> *)

(* cf Collin-Dolex-94 *)

open Algo

let delta=10
let vars = ["path",(At (It,delta)); "par",Nt]
let actions = ["update_path";"compute_parent"]

let (init_vars: neighbor list -> local_env) =
  fun nl -> function
  | "path" -> A (Array.make delta (I (-1)))
  | "par" -> N (try Random.int ((List.length nl)) with _ -> assert false)
  | _ -> raise Not_found
    
(* casting *)
let (a:value -> value array) = function A a -> a | _ -> assert false
let (i:value -> int) = function N i -> i | _ -> assert false

let str_of_array a =
  let l = List.map Algo.value_to_string (Array.to_list a) in
  "["^(String.concat "," l)^"]"

let min_path al =
      match al with
      |  [] -> assert false
      | x::t -> List.fold_left min x t

let (get_paths: neighbor list -> local_env -> value array list) =
  fun nl e ->
    List.mapi (fun alpha_i n -> a (n.lenv "path")) nl 


(* The index of the first negative elt in a starting from i. Returns
   the size of the a if none is found *)
let rec end_of_a a i =
  if i=delta then delta else
  if a.(i) < I 0 then i else end_of_a a (i+1) 
    
    
(* concat and truncate *)
let (concat_path : value array -> int -> value array) =
  fun a alpha ->
    let s = Array.length a in
    let a = Array.copy a in (* ouch! *)
    let last = end_of_a a 1 in
    if last=s then (
      for i = 1 to s-1 do
        a.(i-1) <- a.(i)
      done;
      a.(s-1) <- I alpha
    )
    else
      a.(last) <- I alpha;
    a

let equals_up_to p1 p2 i =
  let res = ref true in
  for j = 0 to i do
    res := !res && p1.(j) = p2.(j) ;
  done;
  !res

let (compute_parent : neighbor list -> value array -> value) =
  fun nl p ->
    (* The parent of a process p in the neighbor which path is a
       subpath of p *)
    let last_p = end_of_a p 1 in
    let paths = List.mapi (fun i n -> a (n.lenv "path"), i) nl in
    let _,parent =
      match 
        List.find_opt
          (fun (npath,_) ->
             let last_n = end_of_a npath 1 in
             last_n = last_p-1 && equals_up_to npath p (last_n-1)
          )
          paths
      with
        Some p -> p
      | None -> [||],(-1) (* may happen before convergence *)
    in
    N parent

let compute_path nl e =
  let paths = get_paths nl e in
  let paths = List.map2 (fun p n -> concat_path p (n.reply ())) paths nl in
  let path = min_path paths in
  path
    
let (enable_f:neighbor list -> local_env -> action list) =
  fun nl e ->
    let path = compute_path nl e in
    if path <> a (e "path") then ["update_path"] else
    if compute_parent nl path <> (e "par") then ["compute_parent"]
    else []

let (step_f : neighbor list -> local_env -> action -> local_env) =
  fun nl e -> 
    function
    | "update_path" ->
      let path = compute_path nl e in
      (
        function
        | "path" -> A path
        | o -> e o
      )
    | "compute_parent" ->
      let path = compute_path nl e in
      (
        function
        | "par"  -> compute_parent nl path
        | o -> e o
      )
    | _ -> assert false

let () =
  let algo_id = "p" in
  Algo.reg_vars      algo_id vars;
  Algo.reg_init_vars algo_id init_vars; 
  Algo.reg_enable    algo_id enable_f;
  Algo.reg_step      algo_id step_f;
  Algo.reg_actions   algo_id actions;
  ()

